/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/windows/graphics.c: Graphics and rendering functionality for
 * Windows.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/debug.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/sysdep/windows/internal.h"

#include "external/opengl-headers/GL/wglext.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Graphics capability structure returned to high-level code.  The display
 * device count and display mode list are filled in at initialization time. */
static SysGraphicsInfo graphics_info = {
    .has_windowed_mode = 1,
    .num_devices = 0,
    .num_modes = 0,
    .modes = NULL,
};

/*---------------------------- General data -----------------------------*/

/* Have we been initialized? */
static uint8_t initted;

/* Should OpenGL debugging be enabled? */
static uint8_t use_opengl_debug;

/* Should we show the mouse pointer? */
static uint8_t show_mouse_pointer;

/* Is the mouse pointer currently shown?  (This is initialized statically
 * rather than in graphics_init() since it reflects persistent system
 * state, though this difference is only relevant when running the test
 * suite.) */
static uint8_t mouse_pointer_visible = 1;

/* Should the mouse pointer be grabbed (confined to the window's rectangle)
 * when a window is visible? */
static uint8_t grab_requested;

/* Is the mouse pointer currently grabbed?  (Again, this reflects
 * persistent system state so we initialize it statically.) */
static uint8_t mouse_grabbed = 0;

/* Window class for creating windows. */
static ATOM window_class;

/* Window style for non-fullscreen windows. */
static const DWORD windowed_style = WS_OVERLAPPEDWINDOW & ~WS_THICKFRAME;

/*-------------------- Display device and mode data ---------------------*/

/* Mapping from our device numbers to Windows device names and associated
 * data.  The length of the array is equal to the value of
 * graphics_info.num_devices. */
static struct {
    char name[32];
    int default_mode;  // Index into device_modes[].
} *devices;

/* Windows DEVMODE structures for each supported mode. */
static DEVMODE *device_modes;

/*------------------ Current display and window state -------------------*/

/* Current display device mode.  This is statically initialized to avoid
 * confusing windows_reset_video_mode() if we error out during setup. */
static int current_mode = -1;

/* Video mode in use when window was iconified.  Used to restore the
 * desired mode when the window is brought back from icon state. */
static int saved_mode = -1;

/* Window handle for current window, or NULL if no window is open. */
static HWND current_window;

/* OpenGL context for the current window. */
static HGLRC wgl_context;

/* Current window size. */
static int window_width, window_height;

/* Window position in windowed mode, used to restore the position when
 * switching out of fullscreen mode. */
static int window_x, window_y;

/* Does the current window have the input focus? */
static uint8_t window_focused;

/* Is the current window in fullscreen mode? */
static uint8_t window_fullscreen;

/* Is the current window minimized? */
static uint8_t window_minimized;

/*--------------------------- Window settings ---------------------------*/

/* Should the window (in windowed mode) be centered when opened? */
static uint8_t center_window;

/* Is fullscreen mode selected (for the next mode change)? */
static uint8_t fullscreen;

/* Is multisampling enabled? */
static uint8_t multisample;

/* Is vertical sync enabled? */
static uint8_t vsync;

/* Display device to use for window creation. */
static int device_to_use;

/* Requested refresh rate. */
static float refresh_rate;

/* Requested depth and stencil sizes. */
static int depth_bits;
static int stencil_bits;

/* Requested number of samples for multisampling. */
static int multisample_samples;

/* Title to use with new windows. */
static char *window_title;

/*------------------------ WGL function pointers ------------------------*/

/* Define function pointers for all wgl*() functions we call.  The
 * non-extension functions could potentially be direct calls, but MSVC
 * doesn't like linking opengl32.lib with our (colliding) dyngl symbols,
 * so we look those up as well to avoid having to link with opengl32. */

static HMODULE opengl32_handle;

static HGLRC (WINAPI *p_wglCreateContext)(HDC dc);
static BOOL (WINAPI *p_wglDeleteContext)(HGLRC context);
static HGLRC (WINAPI *p_wglGetCurrentContext)(void);
static PROC (WINAPI *p_wglGetProcAddress)(const char *name);
static BOOL (WINAPI *p_wglMakeCurrent)(HDC dc, HGLRC context);
static PFNWGLSWAPINTERVALEXTPROC p_wglSwapIntervalEXT;

/* Also define wrappers for standard functions so thread.c can call them. */
HGLRC WINAPI wglCreateContext(HDC dc) {return (*p_wglCreateContext)(dc);}
BOOL WINAPI wglDeleteContext(HGLRC context) {return (*p_wglDeleteContext)(context);}
HGLRC WINAPI wglGetCurrentContext(void) {return (*p_wglGetCurrentContext)();}
PROC WINAPI wglGetProcAddress(const char *name) {return (*p_wglGetProcAddress)(name);}
BOOL WINAPI wglMakeCurrent(HDC dc, HGLRC context) {return (*p_wglMakeCurrent)(dc, context);}


/*--------------------- Local routine declarations ----------------------*/

/**
 * window_proc:  The "window procedure" associated with windows.
 *
 * [Parameters]
 *     hwnd: Window on which the message was received.
 *     uMsg: Message type.
 *     wParam, lParam: Message parameters.
 * [Return value]
 *     Message processing result (the meaning is message-dependent).
 */
static LRESULT CALLBACK window_proc(HWND hwnd, UINT uMsg,
                                    WPARAM wParam, LPARAM lParam);

/**
 * add_video_mode:  Add a video mode to the mode list.
 *
 * [Parameters]
 *     device: Display device index.
 *     mode_info: DEVMODE structure describing the video mode.
 * [Return value]
 *     True on success, false on error (out of memory).
 */
static int add_video_mode(int device, const DEVMODE *mode_info);

/**
 * set_video_mode:  Switch to the given display device mode.
 *
 * [Parameters]
 *     mode: Mode to switch to (index into device_modes[]).
 * [Return value]
 *     True on success, false on error.
 */
static int set_video_mode(int mode);

/**
 * update_mouse_grab:  Enable or disable mouse grabbing via ClipCursor(),
 * depending on whether grabbing has been requested and the current state
 * of the window.
 */
static void update_mouse_grab(void);

/**
 * close_window:  Close the currently open window.
 */
static void close_window(void);

/**
 * setup_opengl:  Initialize the OpenGL subsystem.  This must be called
 * after a display surface has been created.
 *
 * [Parameters]
 *     width, height: Display size, in pixels.
 */
static int setup_opengl(int width, int height);

/**
 * wgl_has_extension:  Return whether the given WGL extension is supported.
 * Requires a window to be open.
 *
 * [Parameters]
 *     name: GLX extension name.
 * [Return value]
 *     True if the extension is supported, false if not.
 */
static int wgl_has_extension(const char *name);

/**
 * wglGetProcAddress_wrapper:  Wrapper for wglGetProcAddress() which falls
 * back to GetProcAddress() on failure (to handle OpenGL 1.0/1.1 functions).
 *
 * [Parameters]
 *     name: Name of function to look up.
 * [Return value]
 *     Function pointer, or NULL if the function is not found.
 */
static void *wglGetProcAddress_wrapper(const char *name);

/**
 * create_wgl_context:  Create a new GL context and set it as the current
 * context.
 *
 * [Parameters]
 *     dc: Device context with which to associate the GL context.
 * [Return value]
 *     Newly created GL context, or zero on error.
 */
static HGLRC create_wgl_context(HDC dc);

/**
 * create_gl_shader_compilation_context:  Create and make current a new
 * GL context for the current thread which can be used to compile shaders.
 *
 * [Return value]
 *     True on success or if the current thread already has a GL context,
 *     false on error.
 */
static int create_gl_shader_compilation_context(void);

/*************************************************************************/
/***************** Interface: Basic graphics operations ******************/
/*************************************************************************/

const SysGraphicsInfo *sys_graphics_init(void)
{
    PRECOND(!initted, return NULL);

    /* Look up basic WGL functions. */
    if (!opengl32_handle) {
        opengl32_handle = LoadLibrary("opengl32.dll");
        if (!opengl32_handle) {
            DLOG("Failed to load opengl32.dll");
            return NULL;
        }
        ASSERT(p_wglCreateContext = (void *)GetProcAddress(
                   opengl32_handle, "wglCreateContext"), return NULL);
        ASSERT(p_wglDeleteContext = (void *)GetProcAddress(
                   opengl32_handle, "wglDeleteContext"), return NULL);
        ASSERT(p_wglGetCurrentContext = (void *)GetProcAddress(
                   opengl32_handle, "wglGetCurrentContext"), return NULL);
        ASSERT(p_wglGetProcAddress = (void *)GetProcAddress(
                   opengl32_handle, "wglGetProcAddress"), return NULL);
        ASSERT(p_wglMakeCurrent = (void *)GetProcAddress(
                   opengl32_handle, "wglMakeCurrent"), return NULL);
    }

    /* Look up connected display devices.  We reorder devices so the
     * primary devices is always device 0. */
    graphics_info.num_devices = 1;
    devices = NULL;
    int got_primary = 0;
    for (int device = 0; ; device++) {
        DISPLAY_DEVICE device_info = {.cb = sizeof(device_info)};
        if (!EnumDisplayDevices(NULL, device, &device_info, 0)) {
            break;
        }
        /* MSDN documents bit 0 as DISPLAY_DEVICE_ACTIVE, but MinGW has
         * DISPLAY_DEVICE_ATTACHED_TO_DESKTOP instead.  Sample code in
         * other MSDN documents also uses ATTACHED_TO_DESKTOP, so maybe
         * ACTIVE is just an older name? */
        if (!(device_info.StateFlags & DISPLAY_DEVICE_ATTACHED_TO_DESKTOP)) {
            continue;
        }
        int index;
        if (device_info.StateFlags & DISPLAY_DEVICE_PRIMARY_DEVICE) {
            ASSERT(!got_primary);
            got_primary = 1;
            index = 0;
        } else {
            index = graphics_info.num_devices++;
        }
        void *new_devices =
            mem_realloc(devices, graphics_info.num_devices * sizeof(*devices),
                        0);
        if (UNLIKELY(!new_devices)) {
            DLOG("Out of memory adding display device %d", index);
            goto error_free_devices;
        }
        devices = new_devices;
        memcpy(devices[index].name, device_info.DeviceName,
               sizeof(devices[index].name));
    }
    if (UNLIKELY(!got_primary)) {
        DLOG("System reports no primary device (OS bug?)");
        graphics_info.num_devices--;
        if (UNLIKELY(graphics_info.num_devices == 0)) {
            DLOG("No display devices found!");
            goto error_return;
        }
        memmove(&devices[0], &devices[1],
                sizeof(*devices) * graphics_info.num_devices);
    }

    /* Look up display modes for each device. */
    graphics_info.num_modes = 0;
    graphics_info.modes = NULL;
    device_modes = NULL;
    for (int device = 0; device < graphics_info.num_devices; device++) {
        DEVMODE default_mode = {.dmSize = sizeof(default_mode)};
        if (UNLIKELY(!EnumDisplaySettings(devices[device].name,
                                          ENUM_REGISTRY_SETTINGS,
                                          &default_mode))) {
            DLOG("Failed to get default mode for device \"%s\" (OS bug?)",
                 devices[device].name);
            mem_clear(&default_mode, sizeof(default_mode));
        }
        devices[device].default_mode = -1;
        const int first_mode = graphics_info.num_modes;
        int got_mode = 0;

        for (int mode = 0; ; mode++) {
            DEVMODE mode_info = {.dmSize = sizeof(mode_info)};
            if (!EnumDisplaySettings(devices[device].name, mode, &mode_info)) {
                break;
            }
            const DWORD required_fields = DM_BITSPERPEL
                                        | DM_PELSWIDTH
                                        | DM_PELSHEIGHT;
            if (UNLIKELY((mode_info.dmFields
                          & required_fields) != required_fields)) {
                DLOG("Mode %d on display %s is missing required fields"
                     " (dmFlags = 0x%X)", mode, devices[device].name,
                     (int)mode_info.dmFields);
                continue;
            }
            if (!(mode_info.dmFields & DM_DISPLAYFREQUENCY)) {
                /* The documentation says this field should always be zero
                 * if the corresponding dmFields bit is not set, but we
                 * clear it anyway just to be sure. */
                mode_info.dmDisplayFrequency = 0;
            }
            if (!(mode_info.dmFields & DM_POSITION)) {
                if (default_mode.dmFields & DM_POSITION) {
                    mode_info.dmFields |= DM_POSITION;
                    mode_info.dmPosition = default_mode.dmPosition;
                } else {
                    mode_info.dmPosition.x = 0;
                    mode_info.dmPosition.y = 0;
                }
            }

            /* Ignore modes with less than 32 bits per pixel. */
            if (mode_info.dmBitsPerPel < 32) {
                continue;
            }

            /* Make sure we could actually change to the mode before we
             * record it. */
            const LONG change_test = ChangeDisplaySettingsEx(
                devices[device].name, &mode_info, NULL,
                CDS_FULLSCREEN | CDS_TEST, NULL);
            if (UNLIKELY(change_test != DISP_CHANGE_SUCCESSFUL)) {
                DLOG("Ignoring mode %d (%dx%dx%d @ %dHz) on display %s"
                     " because ChangeDisplaySettingsEx(CDS_TEST) failed"
                     " with code %ld", mode, (int)mode_info.dmPelsWidth,
                     (int)mode_info.dmPelsHeight, (int)mode_info.dmBitsPerPel,
                     (int)mode_info.dmDisplayFrequency, devices[device].name,
                     change_test);
                continue;
            }

            /* This mode looks good, so add it to the arrays. */
            got_mode = 1;
            if (UNLIKELY(!add_video_mode(device, &mode_info))) {
                goto error_free_sizes;
            }
            if (devices[device].default_mode < 0
             && mode_info.dmDisplayFrequency == default_mode.dmDisplayFrequency
             && mode_info.dmPelsWidth == default_mode.dmPelsWidth
             && mode_info.dmPelsHeight == default_mode.dmPelsHeight
             && mode_info.dmBitsPerPel == default_mode.dmBitsPerPel) {
                devices[device].default_mode = graphics_info.num_modes - 1;
            }
        }  // for (int mode = 0; ; mode++)

        if (devices[device].default_mode < 0
         && default_mode.dmBitsPerPel >= 32) {
            DLOG("Failed to find default mode for device %s, adding to list:"
                 " %dx%dx%d @ %dHz", devices[device].name,
                 (int)default_mode.dmPelsWidth,
                 (int)default_mode.dmPelsHeight,
                 (int)default_mode.dmBitsPerPel,
                 (int)default_mode.dmDisplayFrequency);
            got_mode = 1;
            if (UNLIKELY(!add_video_mode(device, &default_mode))) {
                goto error_free_sizes;
            }
            devices[device].default_mode = graphics_info.num_modes - 1;
        }
        if (UNLIKELY(!got_mode)) {
            DLOG("Failed to get any modes for device %s, removing from list",
                 devices[device].name);
            graphics_info.num_devices--;
            if (UNLIKELY(graphics_info.num_devices == 0)) {
                DLOG("No devices left!");
                goto error_free_devices;
            }
            memmove(&devices[device], &devices[device+1],
                    sizeof(*devices) * (graphics_info.num_devices - device));
            device--;
        } else if (UNLIKELY(devices[device].default_mode < 0)) {
            DLOG("Failed to find default mode for device %s, using first"
                 " mode in list: %dx%dx%d @ %dHz", devices[device].name,
                 (int)device_modes[first_mode].dmPelsWidth,
                 (int)device_modes[first_mode].dmPelsHeight,
                 (int)device_modes[first_mode].dmBitsPerPel,
                 (int)device_modes[first_mode].dmDisplayFrequency);
            devices[device].default_mode = first_mode;
        }
    }

    /* Set up a window class to use when creating windows. */
    WNDCLASSEX class_def = {
        .cbSize = sizeof(class_def),
        .style = 0,
        .lpfnWndProc = window_proc,
        .cbClsExtra = 0,
        .cbWndExtra = 0,
        .hInstance = GetModuleHandle(NULL),
        .hIcon = NULL,
        .hCursor = NULL,
        .hbrBackground = NULL,
        .lpszMenuName = NULL,
        .lpszClassName = "SILWindowClass",
        .hIconSm = NULL,
    };
#ifdef HAVE_DEFAULT_ICON
    /* Pass LR_SHARED so the system automatically cleans up after us. */
    class_def.hIcon = LoadImage(GetModuleHandle(NULL), "DefaultIcon",
                                IMAGE_ICON, 0, 0, LR_DEFAULTSIZE | LR_SHARED);
    if (!class_def.hIcon) {
        DLOG("Failed to load default icon: %s",
             windows_strerror(GetLastError()));
    }
#endif
    window_class = RegisterClassEx(&class_def);
    if (!window_class) {
        DLOG("Failed to create window class");
        goto error_free_sizes;
    }

    /* Initialize other internal state and return. */
    center_window = 0;
    current_mode = devices[0].default_mode;
    current_window = NULL;
    depth_bits = 16;
    device_to_use = 0;
    fullscreen = 0;
    grab_requested = 0;
    multisample = 0;
    multisample_samples = 1;
    refresh_rate = 0;
    show_mouse_pointer = 0;
    stencil_bits = 0;
    use_opengl_debug = 0;
    vsync = 1;
    wgl_context = NULL;
    window_focused = 0;

    initted = 1;
    return &graphics_info;

  error_free_sizes:
    mem_free((void *)graphics_info.modes);
    graphics_info.modes = NULL;
    graphics_info.num_modes = 0;
    mem_free(device_modes);
    device_modes = NULL;
  error_free_devices:
    mem_free(devices);
    devices = NULL;
    graphics_info.num_devices = 0;
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_cleanup(void)
{
    PRECOND(initted, return);

    /* Reset the video mode before closing the window so the monitor is
     * already switching modes when the window disappears. */
    windows_reset_video_mode();

    if (current_window) {
        close_window();
    }
    mem_free(window_title);
    window_title = NULL;

    UnregisterClass((void *)(uintptr_t)window_class, GetModuleHandle(NULL));
    window_class = 0;

    mem_free((void *)graphics_info.modes);
    graphics_info.modes = NULL;
    graphics_info.num_modes = 0;
    mem_free(device_modes);
    device_modes = NULL;
    mem_free(devices);
    devices = NULL;
    graphics_info.num_devices = 0;

    initted = 0;
}

/*-----------------------------------------------------------------------*/

int sys_graphics_device_width(void)
{
    const int device = graphics_info.modes[current_mode].device;
    const int mode = devices[device].default_mode;
    return graphics_info.modes[mode].width;
}

/*-----------------------------------------------------------------------*/

int sys_graphics_device_height(void)
{
    const int device = graphics_info.modes[current_mode].device;
    const int mode = devices[device].default_mode;
    return graphics_info.modes[mode].height;
}

/*-----------------------------------------------------------------------*/

int sys_graphics_set_display_attr(const char *name, va_list args)
{
    if (strcmp(name, "center_window") == 0) {
        const int value = va_arg(args, int);
        center_window = (value != 0);
        return 1;
    }

    if (strcmp(name, "depth_bits") == 0) {
        const int value = va_arg(args, int);
        if (value < 0) {
            DLOG("Invalid value for attribute depth_bits: %d", value);
            return 0;
        }
        depth_bits = value;
        opengl_set_depth_bits(depth_bits);
        return 1;
    }

    if (strcmp(name, "device") == 0) {
        const int value = va_arg(args, int);
        if (value < 0 || value > graphics_info.num_devices) {
            DLOG("Invalid value for attribute device: %d", value);
            return 0;
        }
        device_to_use = value;
        return 1;
    }

    if (strcmp(name, "multisample") == 0) {
        const int value = va_arg(args, int);
        if (value <= 0) {
            DLOG("Invalid value for attribute multisample: %d", value);
            return 0;
        }
        multisample = (value > 1);
        multisample_samples = value;
        return 1;
    }

    if (strcmp(name, "opengl_debug") == 0) {
        use_opengl_debug = (va_arg(args, int) != 0);
        /* HACK: Some Intel drivers (at least) claim to support
         * ARB_debug_output but don't implement wglCreateContextAttribsARB()
         * to enable the debug bit, and then ignore all debug messages
         * (which is technically a compliant implementation since we didn't
         * set the context's debug bit -- never mind that the driver
         * doesn't give us a way to set it in the first place).  Reject
         * attempts to enable debugging if the driver doesn't provide the
         * wglCreateContextAttribsARB() function. */
        if (use_opengl_debug
            && !GetProcAddress(GetModuleHandle("opengl32.dll"),
                               "wglGetContextAttribsARB")) {
            DLOG("Driver doesn't support wglGetContextAttribsARB(), can't"
                 " enable OpenGL debugging");
            return 0;
        }
        /* We don't (can't) look up GL functions until after opening a
         * window, so it's not necessarily safe to call opengl_enable_debug()
         * here. */
        if (current_window) {
            opengl_enable_debug(use_opengl_debug);
        }
        return 1;
    }

    if (strcmp(name, "refresh_rate") == 0) {
        const float value = (float)va_arg(args, double);
        if (!(value >= 0)) {
            DLOG("Invalid value for attribute refresh_rate: %g", value);
            return 0;
        }
        refresh_rate = value;
        return 1;
    }

    if (strcmp(name, "stencil_bits") == 0) {
        const int value = va_arg(args, int);
        if (value < 0) {
            DLOG("Invalid value for attribute stencil_bits: %d", value);
            return 0;
        }
        stencil_bits = value;
        opengl_set_stencil_bits(stencil_bits);
        return 1;
    }

    if (strcmp(name, "vsync") == 0) {
        vsync = (va_arg(args, int) != 0);
        if (current_window && p_wglSwapIntervalEXT) {
            (*p_wglSwapIntervalEXT)(vsync);
        }
        return 1;
    }

    if (strcmp(name, "window") == 0) {
        fullscreen = (va_arg(args, int) == 0);
        return 1;
    }

    return 0;
}

/*-----------------------------------------------------------------------*/

GraphicsError sys_graphics_set_display_mode(int width, int height)
{
    GraphicsError error = GRAPHICS_ERROR_UNKNOWN;

    /* Look up the device mode corresponding to the requested device and
     * display size. */
    int fullscreen_mode = -1;
    if (fullscreen) {
        for (int i = 0; i < graphics_info.num_modes; i++) {
            if (graphics_info.modes[i].width == width
             && graphics_info.modes[i].height == height
             && graphics_info.modes[i].device == device_to_use) {
                if (fullscreen_mode < 0) {
                    fullscreen_mode = i;
                } else if (refresh_rate > 0) {
                    if (fabsf(graphics_info.modes[i].refresh - refresh_rate)
                        < fabsf(graphics_info.modes[fullscreen_mode].refresh
                                - refresh_rate))
                    {
                        fullscreen_mode = i;
                    }
                } else {  // refresh_rate == 0
                    if (graphics_info.modes[i].refresh
                        > graphics_info.modes[fullscreen_mode].refresh)
                    {
                        fullscreen_mode = i;
                    }
                }
            }
        }
        if (fullscreen_mode < 0) {
            DLOG("No video mode matching %dx%d on device %d", width, height,
                 device_to_use);
            error = GRAPHICS_ERROR_MODE_NOT_SUPPORTED;
            goto error_return;
        }
    }

    /* See if we can toggle fullscreen or change size without closing the
     * window. */
    if (current_window
     && device_to_use == graphics_info.modes[current_mode].device) {
        GLint window_depth = -1;
        GLint window_buffers = -1;
        GLint window_samples = -1;
        GLint window_stencil = -1;
        glGetIntegerv(GL_DEPTH_BITS, &window_depth);
        glGetIntegerv(GL_SAMPLE_BUFFERS, &window_buffers);
        glGetIntegerv(GL_SAMPLES, &window_samples);
        glGetIntegerv(GL_STENCIL_BITS, &window_stencil);
        if (window_depth >= depth_bits
         && (!multisample
             || (window_buffers && window_samples >= multisample_samples))
         && window_stencil >= stencil_bits)
        {
            if (width == window_width && height == window_height) {
                /* Window size is unchanged -- we're just (possibly)
                 * switching into or out of fullscreen mode. */
                if (fullscreen == window_fullscreen) {
                    /* No change at all! */
                    return GRAPHICS_ERROR_SUCCESS;
                } else {
                    if (!fullscreen) {
                        /* Save the current window position for restoring
                         * from fullscreen mode. */
                        RECT rect;
                        if (UNLIKELY(!GetWindowRect(current_window, &rect))) {
                            DLOG("Failed to save window position: %s",
                                 windows_strerror(GetLastError()));
                            window_x = window_y = 0;
                        } else {
                            window_x = rect.left
                                - device_modes[current_mode].dmPosition.x;
                            window_y = rect.top
                                - device_modes[current_mode].dmPosition.y;
                        }
                    }
                    /* Switch modes if necessary. */
                    if (!fullscreen) {
                        windows_reset_video_mode();
                    } else {
                        set_video_mode(fullscreen_mode);
                    }
                    /* Toggle the window borders on or off as appropriate.
                     * (Whose clever idea was it to have a public API
                     * which is essentially just peek/poke into a private
                     * structure?  Good grief...) */
                    DWORD style = GetWindowLong(current_window, GWL_STYLE);
                    style &= ~(windowed_style | WS_POPUP);
                    style |= fullscreen ? WS_POPUP : windowed_style;
                    SetWindowLong(current_window, GWL_STYLE, style);
                    /* Move and resize the window as appropriate. */
                    int x, y, w, h;
                    if (fullscreen) {
                        x = device_modes[fullscreen_mode].dmPosition.x;
                        y = device_modes[fullscreen_mode].dmPosition.y;
                        w = width;
                        h = height;
                    } else {
                        RECT rect = {.left = 0, .top = 0, .right = window_width,
                                     .bottom = window_height};
                        if (UNLIKELY(!AdjustWindowRectEx(&rect, style, FALSE,
                                                         0))) {
                            DLOG("AdjustWindowRectEx() failed for fullscreen"
                                 " toggle: %s",
                                 windows_strerror(GetLastError()));
                        }
                        x = window_x;
                        y = window_y;
                        w = rect.right - rect.left;
                        h = rect.bottom - rect.top;
                    }
                    if (SetWindowPos(current_window,
                                     fullscreen ? HWND_TOPMOST : HWND_NOTOPMOST,
                                     x, y, w, h, SWP_NOCOPYBITS)) {
                        window_fullscreen = !window_fullscreen;
                        return GRAPHICS_ERROR_SUCCESS;
                    } else {
                        DLOG("SetWindowPos() failed for fullscreen toggle: %s",
                             windows_strerror(GetLastError()));
                    }
                }
            } else if (!fullscreen && !window_fullscreen) {
                /* We're resizing the window in windowed mode. */
                const DWORD style = GetWindowLong(current_window, GWL_STYLE);
                RECT rect = {.left = 0, .top = 0, .right = width,
                             .bottom = height};
                if (UNLIKELY(!AdjustWindowRectEx(&rect, style, FALSE, 0))) {
                    DLOG("AdjustWindowRectEx() failed for window resize: %s",
                         windows_strerror(GetLastError()));
                }
                /* We have to change window_width and window_height before
                 * calling SetWindowPos(), since that function sends a
                 * WM_GETMINMAXINFO to the window to bound the size, which
                 * in turn reads from window_{width,height}. */
                const int old_width = window_width;
                const int old_height = window_height;
                window_width = width;
                window_height = height;
                if (SetWindowPos(current_window, HWND_NOTOPMOST, 0, 0,
                                 rect.right - rect.left, rect.bottom - rect.top,
                                 SWP_NOMOVE | SWP_NOACTIVATE)) {
                    opengl_set_display_size(width, height);
                    return GRAPHICS_ERROR_SUCCESS;
                } else {
                    DLOG("SetWindowPos() failed for window resize: %s",
                         windows_strerror(GetLastError()));
                    window_width = old_width;
                    window_height = old_height;
                }
            }
        }
    }

    /* If we already had a window open, save OpenGL state to memory, then
     * destroy the window and associated GL context. */
    const int window_was_open = (current_window != NULL);
    if (window_was_open) {
        close_window();
    }

    /* If we're changing display devices, restore the old device's mode
     * and select the new device's default mode as current (to ensure the
     * centering calculation below works correctly). */
    if (graphics_info.modes[current_mode].device != device_to_use) {
        ASSERT(set_video_mode(devices[device_to_use].default_mode));
    }

    /* Determine the initial window position and total size. */
    RECT new_rect = {.left = device_modes[current_mode].dmPosition.x,
                     .top = device_modes[current_mode].dmPosition.y,
                     .right = device_modes[current_mode].dmPosition.x+width,
                     .bottom = device_modes[current_mode].dmPosition.y+height};
    /* Windows doesn't allow us to say "default position on non-default
     * monitor", so force centering if the window is supposed to go to an
     * alternate monitor. */
    const int do_center = center_window || device_to_use != 0;
    if (!fullscreen) {
        if (do_center) {
            const int device_width = graphics_info.modes[current_mode].width;
            const int device_height = graphics_info.modes[current_mode].height;
            const int x_offset = (device_width - width) / 2;
            const int y_offset = (device_height - height) / 2;
            new_rect.left += x_offset;
            new_rect.top += y_offset;
            new_rect.right += x_offset;
            new_rect.bottom += y_offset;
        }
        AdjustWindowRectEx(&new_rect, windowed_style, FALSE, 0);
    }

    /* Change video modes if creating a fullscreen window.  Note that
     * Windows ignores the initial window mode (normal/minimized/maximized)
     * for WS_POPUP windows, so we change the video mode regardless of what
     * was specified in the STARTUPINFO structure. */
    if (fullscreen) {
        set_video_mode(fullscreen_mode);
    }

    /* Actually create the window and associated GL context. */
    window_width = width;
    window_height = height;
    window_focused = 1;
    window_fullscreen = (fullscreen != 0);
    window_minimized = 0;
    current_window = CreateWindow(
        (void *)(uintptr_t)window_class, window_title,
        fullscreen ? WS_POPUP : windowed_style,
        fullscreen || do_center ? new_rect.left : (LONG)CW_USEDEFAULT,
        fullscreen || do_center ? new_rect.top : (LONG)CW_USEDEFAULT,
        new_rect.right - new_rect.left, new_rect.bottom - new_rect.top,
        NULL, NULL, GetModuleHandle(NULL), NULL);
    if (UNLIKELY(!current_window)) {
        DLOG("CreateWindow() failed: %s", windows_strerror(GetLastError()));
        error = GRAPHICS_ERROR_MODE_NOT_SUPPORTED;
        goto error_return;
    }
    HDC dc = GetDC(current_window);
    ASSERT(dc);  // Should never fail.
    PIXELFORMATDESCRIPTOR pfd = {
        .nSize           = sizeof(pfd),
        .nVersion        = 1,
        .dwFlags         = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL
                           | PFD_DOUBLEBUFFER
                           | (depth_bits == 0 ? PFD_DEPTH_DONTCARE : 0),
        .iPixelType      = PFD_TYPE_RGBA,
        .cColorBits      = 24,
        .cRedBits        = 8,
        .cRedShift       = 0,
        .cGreenBits      = 8,
        .cGreenShift     = 0,
        .cBlueBits       = 8,
        .cBlueShift      = 0,
        .cAlphaBits      = 0,
        .cAlphaShift     = 0,
        .cAccumBits      = 0,
        .cAccumRedBits   = 0,
        .cAccumGreenBits = 0,
        .cAccumBlueBits  = 0,
        .cAccumAlphaBits = 0,
        .cDepthBits      = depth_bits,
        .cStencilBits    = stencil_bits,
        .cAuxBuffers     = 0,
        .iLayerType      = 0,
        .bReserved       = 0,
        .dwLayerMask     = 0,
        .dwVisibleMask   = 0,
        .dwDamageMask    = 0,
    };
    int pixel_format = ChoosePixelFormat(dc, &pfd);
    if (!pixel_format) {
        DLOG("Failed to choose a pixel format: %s",
             windows_strerror(GetLastError()));
        error = GRAPHICS_ERROR_MODE_NOT_SUPPORTED;
        goto error_destroy_window;
    }
    if (UNLIKELY(!DescribePixelFormat(dc, pixel_format, sizeof(pfd), &pfd))) {
        DLOG("DescribePixelFormat() failed: %s",
             windows_strerror(GetLastError()));
        goto error_destroy_window;
    }
    if (pfd.cRedBits < 8
     || pfd.cGreenBits < 8
     || pfd.cBlueBits < 8
     || pfd.cDepthBits < depth_bits
     || pfd.cStencilBits < stencil_bits) {
        DLOG("No matching pixel formats found");
        error = GRAPHICS_ERROR_MODE_NOT_SUPPORTED;
        goto error_destroy_window;
    }
    if (UNLIKELY(!SetPixelFormat(dc, pixel_format, &pfd))) {
        DLOG("SetPixelFormat() failed: %s", windows_strerror(GetLastError()));
        goto error_destroy_window;
    }
    wgl_context = create_wgl_context(dc);
    if (UNLIKELY(!wgl_context)) {
        DLOG("Failed to create GL context");
        goto error_destroy_window;
    }

    /* If multisampling is requested, we have to do the context creation
     * all over again because we need wglChoosePixelFormatARB() for that,
     * but wglChoosePixelFormat() is only available via dynamic lookup,
     * and we can't look up functions until we have a context. */
    if (multisample) {
        PFNWGLCHOOSEPIXELFORMATARBPROC p_wglChoosePixelFormatARB;
        p_wglChoosePixelFormatARB =
            (void *)wglGetProcAddress("wglChoosePixelFormatARB");
        if (p_wglChoosePixelFormatARB) {
            const int attributes[] = {
                WGL_DRAW_TO_WINDOW_ARB, TRUE,
                WGL_SUPPORT_OPENGL_ARB, TRUE,
                WGL_DOUBLE_BUFFER_ARB,  TRUE,
                WGL_PIXEL_TYPE_ARB,     WGL_TYPE_RGBA_ARB,
                WGL_COLOR_BITS_ARB,     24,
                WGL_RED_BITS_ARB,       8,
                WGL_GREEN_BITS_ARB,     8,
                WGL_BLUE_BITS_ARB,      8,
                WGL_DEPTH_BITS_ARB,     depth_bits,
                WGL_STENCIL_BITS_ARB,   stencil_bits,
                WGL_SAMPLE_BUFFERS_ARB, 1,
                WGL_SAMPLES_ARB,        multisample_samples,
                0
            };
            UINT formats_found = 0;
            if (!(*p_wglChoosePixelFormatARB)(
                    dc, attributes, NULL, 1, &pixel_format,
                    &formats_found)) {
                DLOG("wglChoosePixelFormatARB() failed");
                error = GRAPHICS_ERROR_MODE_NOT_SUPPORTED;
                goto error_close_context;
            } else if (formats_found == 0) {
                DLOG("No matching multisample formats found");
                error = GRAPHICS_ERROR_MODE_NOT_SUPPORTED;
                goto error_close_context;
            } else {
                wglMakeCurrent(dc, NULL);
                wglDeleteContext(wgl_context);
                wgl_context = NULL;
                /* Windows doesn't let us change the pixel format of a window,
                 * so we have to destroy and recreate the window as well. */
                ReleaseDC(current_window, dc);
                DestroyWindow(current_window);
                current_window = CreateWindow(
                    (void *)(uintptr_t)window_class, window_title,
                    fullscreen ? WS_POPUP : windowed_style,
                    fullscreen || center_window ? new_rect.left : (LONG)CW_USEDEFAULT,
                    fullscreen || center_window ? new_rect.top : (LONG)CW_USEDEFAULT,
                    new_rect.right - new_rect.left,
                    new_rect.bottom - new_rect.top,
                    NULL, NULL, GetModuleHandle(NULL), NULL);
                if (UNLIKELY(!current_window)) {
                    DLOG("CreateWindow() failed (multisample): %s",
                         windows_strerror(GetLastError()));
                    error = GRAPHICS_ERROR_MODE_NOT_SUPPORTED;
                    goto error_return;
                }
                dc = GetDC(current_window);
                ASSERT(dc);
                if (UNLIKELY(!SetPixelFormat(dc, pixel_format, &pfd))) {
                    DLOG("SetPixelFormat() failed (multisample): %s",
                         windows_strerror(GetLastError()));
                    goto error_destroy_window;
                }
                wgl_context = create_wgl_context(dc);
                if (UNLIKELY(!wgl_context)) {
                    DLOG("Failed to create GL context (multisample)");
                    goto error_destroy_window;
                }
            }
        } else {
            DLOG("wglChoosePixelFormatARB() unavailable, can't enable"
                 " multisampling");
            error = GRAPHICS_ERROR_MODE_NOT_SUPPORTED;
            goto error_close_context;
        }
    }

    /* OpenGL function pointers in Windows are context-specific, so we
     * have to reinitialize them each time around. */
    if (wgl_has_extension("WGL_EXT_swap_control")) {
        p_wglSwapIntervalEXT =
            (void *)wglGetProcAddress("wglGetSwapIntervalEXT");
    } else {
        p_wglSwapIntervalEXT = NULL;
    }
    opengl_lookup_functions(wglGetProcAddress_wrapper);

    /* Set up OpenGL now that we have a context to work with. */
    if (UNLIKELY(!setup_opengl(width, height))) {
        return GRAPHICS_ERROR_OPENGL_TOO_OLD;
    }

    /* Enable multisampling if requested. */
    if (multisample) {
        glEnable(GL_MULTISAMPLE);
    } else {
        glDisable(GL_MULTISAMPLE);
    }

    /* Display the window, hide the pointer if requested, and return. */
    ShowWindow(current_window, SW_SHOWNORMAL);
    ReleaseDC(current_window, dc);
    windows_show_mouse_pointer(-1);
    return window_was_open ? GRAPHICS_ERROR_STATE_LOST
                           : GRAPHICS_ERROR_SUCCESS;

  error_close_context:
    wglMakeCurrent(dc, NULL);
    wglDeleteContext(wgl_context);
    wgl_context = NULL;
  error_destroy_window:
    ReleaseDC(current_window, dc);
    DestroyWindow(current_window);
    current_window = NULL;
  error_return:
    return error;
}

/*-----------------------------------------------------------------------*/

int sys_graphics_display_is_window(void)
{
    return current_window && !window_fullscreen;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_set_window_title(const char *title)
{
    mem_free(window_title);
    window_title = mem_strdup(title, 0);
    if (UNLIKELY(!window_title)) {
        DLOG("No memory to copy window title");
    }
    if (current_window) {
        SetWindowText(current_window, title);
    }
}

/*-----------------------------------------------------------------------*/

void sys_graphics_set_window_icon(SysTexture *texture)
{
    const int width = sys_texture_width(texture);
    const int height = sys_texture_height(texture);
    const int large_icon = (width > 16 || height > 16);

    uint8_t *pixels = sys_texture_lock(texture, SYS_TEXTURE_LOCK_NORMAL,
                                       0, 0, width, height);
    if (!pixels) {
        DLOG("Failed to lock texture for icon");
        return;
    }

    /* Windows wants the image in BMP format. */
    const int stride = width * 4;
    const int bmp_size = sizeof(BITMAPINFOHEADER) + height * stride;
    uint8_t *bmp = mem_alloc(bmp_size, 0, MEM_ALLOC_TEMP);
    if (UNLIKELY(!bmp)) {
        DLOG("No memory for temporary copy of icon (%d bytes)", bmp_size);
        sys_texture_unlock(texture, 0);
        return;
    }
    BITMAPINFOHEADER *bmp_header = (BITMAPINFOHEADER *)bmp;
    mem_clear(bmp_header, sizeof(*bmp_header));
    bmp_header->biSize = sizeof(*bmp_header);
    bmp_header->biWidth = width;
    /* Windows icon bitmaps are double height, with the second half of
     * the buffer (the upper half of the bitmap) containing an AND mask
     * for use with monochrome displays.  Those things don't exist
     * anymore, but we still need to double the height as recorded in
     * the bitmap header or Windows will take the bottom half of our icon
     * and stretch it vertically. */
    bmp_header->biHeight = height * 2;
    bmp_header->biPlanes = 1;
    bmp_header->biBitCount = 32;
    bmp_header->biCompression = BI_RGB;
    bmp_header->biSizeImage = height * stride;
    /* Vertically flip the image and swap RGB order to match BMP format.
     * (Windows XP, at least, doesn't accept bitmaps in top-down order.) */
    for (int y = 0; y < height; y++) {
        const uint8_t *src = pixels + ((height-1)-y) * stride;
        uint8_t *dest = bmp + sizeof(BITMAPINFOHEADER) + y * stride;
        for (int x = 0; x < width; x++) {
            dest[x*4+0] = src[x*4+2];
            dest[x*4+1] = src[x*4+1];
            dest[x*4+2] = src[x*4+0];
            dest[x*4+3] = src[x*4+3];
        }
    }
    sys_texture_unlock(texture, 0);

    /* 0x00030000 is a magic value defined in the CreateIconFromResourceEx()
     * documentation. */
    HANDLE icon = CreateIconFromResourceEx(
        bmp, bmp_size, TRUE, 0x00030000, 0, 0, LR_SHARED);
    mem_free(bmp);
    if (UNLIKELY(!icon)) {
        DLOG("Failed to create icon resource: %s",
             windows_strerror(GetLastError()));
        return;
    }

    SendMessage(current_window, WM_SETICON, ICON_SMALL, (LPARAM)icon);
    if (large_icon) {
        SendMessage(current_window, WM_SETICON, ICON_BIG, (LPARAM)icon);
    }
}

/*-----------------------------------------------------------------------*/

void sys_graphics_show_mouse_pointer(int on)
{
    show_mouse_pointer = (on != 0);
    windows_show_mouse_pointer(-1);
}

/*-----------------------------------------------------------------------*/

int sys_graphics_get_mouse_pointer_state(void)
{
    return show_mouse_pointer;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_get_frame_period(int *numerator_ret, int *denominator_ret)
{
    if (device_modes[current_mode].dmDisplayFrequency) {
        *numerator_ret = 1;
        *denominator_ret = device_modes[current_mode].dmDisplayFrequency;
    } else {
        *numerator_ret = 0;
        *denominator_ret = 1;
    }
}

/*-----------------------------------------------------------------------*/

int sys_graphics_has_focus(void)
{
    return window_focused;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_start_frame(int *width_ret, int *height_ret)
{
    *width_ret = window_width;
    *height_ret = window_height;

#ifdef DEBUG
    debug_record_cpu_phase(DEBUG_CPU_PROCESS_END);
    debug_record_cpu_phase(DEBUG_CPU_GPU_WAIT_START);
    opengl_sync();
    debug_record_cpu_phase(DEBUG_CPU_GPU_WAIT_END);
#endif

#ifdef DEBUG
    debug_record_cpu_phase(DEBUG_CPU_RENDER_START);
#endif
    opengl_start_frame();
}

/*-----------------------------------------------------------------------*/

void sys_graphics_finish_frame(void)
{
#ifdef DEBUG
    debug_record_cpu_phase(DEBUG_CPU_RENDER_END);
#endif
    HDC dc = GetDC(current_window);
    SwapBuffers(dc);
    ReleaseDC(current_window, dc);
#ifdef DEBUG
    debug_record_cpu_phase(DEBUG_CPU_PROCESS_START);
#endif
}

/*-----------------------------------------------------------------------*/

void sys_graphics_sync(int flush)
{
    opengl_sync();
    if (flush) {
        opengl_free_dead_resources(1);
    }
}

/*************************************************************************/
/****************** Windows-internal exported routines *******************/
/*************************************************************************/

void windows_close_window(void)
{
    if (current_window) {
        close_window();
    }
}

/*-----------------------------------------------------------------------*/

void windows_reset_video_mode(void)
{
    if (current_mode >= 0) {
        const int default_mode =
            devices[graphics_info.modes[current_mode].device].default_mode;
        if (current_mode != default_mode) {
            set_video_mode(default_mode);
        }
    }
}

/*-----------------------------------------------------------------------*/

void windows_set_mouse_grab(int grab)
{
    grab_requested = (grab != 0);
    update_mouse_grab();
}

/*-----------------------------------------------------------------------*/

void windows_show_mouse_pointer(int override)
{
    const int on = (override >= 0) ? override : show_mouse_pointer;

    /* ShowCursor() acts like a counter rather than a boolean flag, so we
     * have to be careful to only call it when actually changing the state
     * of the cursor. */
    if (on && !mouse_pointer_visible) {
        mouse_pointer_visible = 1;
        ShowCursor(TRUE);
    } else if (!on && mouse_pointer_visible) {
        mouse_pointer_visible = 0;
        ShowCursor(FALSE);
    }
}

/*-----------------------------------------------------------------------*/

HGLRC windows_wgl_context(void)
{
    return wgl_context;
}

/*-----------------------------------------------------------------------*/

HWND windows_window(void)
{
    return current_window;
}

/*-----------------------------------------------------------------------*/

const char *windows_window_title(void)
{
    return window_title;
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static LRESULT CALLBACK window_proc(HWND hwnd, UINT uMsg,
                                    WPARAM wParam, LPARAM lParam)
{
    if (uMsg == WM_GETMINMAXINFO) {
        /* Prevent window size changes. */
        RECT rect = {.left = 0, .top = 0,
                     .right = window_width, .bottom = window_height};
        AdjustWindowRectEx(&rect, windowed_style, FALSE, 0);
        MINMAXINFO *info = (MINMAXINFO *)lParam;
        info->ptMinTrackSize.x = rect.right - rect.left;
        info->ptMinTrackSize.y = rect.bottom - rect.top;
        info->ptMaxTrackSize.x = rect.right - rect.left;
        info->ptMaxTrackSize.y = rect.bottom - rect.top;
        return FALSE;

    } else if (uMsg == WM_KILLFOCUS) {
        /* We get all three of WM_ACTIVATEAPP, WM_ACTIVATE, and WM_SETFOCUS
         * when a window is activated, and similarly (but with WM_KILLFOCUS)
         * when a window is deactivated.  We use WM_{SET,KILL}FOCUS as the
         * trigger for changing focus since they reflect where keyboard
         * input will go, though in practice it may not make any difference. */
        window_focused = 0;
        update_mouse_grab();
        return DefWindowProc(hwnd, uMsg, wParam, lParam);

    } else if (uMsg == WM_SETCURSOR) {
        /* We don't want to hide the cursor in the window's title bar,
         * so override the current ShowCursor() setting in that case.
         * (courtesy stackoverflow://5629613) */
        if (LOWORD(lParam) == HTCLIENT) {
            windows_show_mouse_pointer(-1);
        } else {
            windows_show_mouse_pointer(1);
        }
        return DefWindowProc(hwnd, uMsg, wParam, lParam);

    } else if (uMsg == WM_SETFOCUS) {
        window_focused = 1;
        update_mouse_grab();
        return DefWindowProc(hwnd, uMsg, wParam, lParam);

    } else if (uMsg == WM_SYSCOMMAND && (wParam & 0xFFF0) == SC_MINIMIZE) {
        if (!window_minimized) {
            if (window_fullscreen) {
                saved_mode = current_mode;
                windows_reset_video_mode();
            }
            update_mouse_grab();
            window_minimized = 1;
        }
        return DefWindowProc(hwnd, uMsg, wParam, lParam);

    } else if (uMsg == WM_SYSCOMMAND && (wParam & 0xFFF0) == SC_RESTORE) {
        if (window_minimized) {
            if (window_fullscreen) {
                if (UNLIKELY(!set_video_mode(saved_mode))) {
                    /* Try to recover by switching to windowed mode. */
                    const int saved_fullscreen = fullscreen;
                    fullscreen = 0;
                    sys_graphics_set_display_mode(window_width, window_height);
                    fullscreen = saved_fullscreen;
                }
            }
            window_minimized = 0;
        }
        return DefWindowProc(hwnd, uMsg, wParam, lParam);

    } else if (uMsg == WM_SYSCOMMAND && (wParam & 0xFFF0) == SC_MAXIMIZE) {
        if (window_minimized) {
            DefWindowProc(hwnd, WM_SYSCOMMAND, SC_RESTORE, 0);
            window_minimized = 0;
        }
        return FALSE;

    } else if (uMsg == WM_UNICHAR && wParam == UNICODE_NOCHAR) {
        return TRUE;  // Declare that we support WM_UNICHAR.

    } else if (windows_handle_input_message(hwnd, uMsg, wParam, lParam)) {
        return FALSE;

    } else {
        return DefWindowProc(hwnd, uMsg, wParam, lParam);
    }
}

/*-----------------------------------------------------------------------*/

static int add_video_mode(int device, const DEVMODE *mode_info)
{
    const int new_num_modes = graphics_info.num_modes + 1;
    void *new_device_modes = mem_realloc(
        device_modes, sizeof(*device_modes) * new_num_modes, 0);
    if (UNLIKELY(!new_device_modes)) {
        DLOG("No memory to expand device mode list to %d entries",
             new_num_modes);
        return 0;
    }
    device_modes = new_device_modes;
    GraphicsDisplayModeEntry *modes = mem_realloc(
        (GraphicsDisplayModeEntry *)graphics_info.modes,
        sizeof(*graphics_info.modes) * new_num_modes, 0);
    if (UNLIKELY(!modes)) {
        DLOG("No memory to expand mode list to %d entries", new_num_modes);
        return 0;
    }
    graphics_info.modes = modes;
    const int index = graphics_info.num_modes++;
    device_modes[index] = *mode_info;
    modes[index].device = device;
    modes[index].device_name = devices[device].name;
    modes[index].width = mode_info->dmPelsWidth;
    modes[index].height = mode_info->dmPelsHeight;
    modes[index].refresh = mode_info->dmDisplayFrequency;
    return 1;
}

/*-----------------------------------------------------------------------*/

static int set_video_mode(int mode)
{
    PRECOND(mode >= 0 && mode < graphics_info.num_modes, return 0);

    const int device = graphics_info.modes[mode].device;
    if (device != graphics_info.modes[current_mode].device) {
        windows_reset_video_mode();
        current_mode = devices[device].default_mode;
    }

    if (mode == current_mode) {
        return 1;
    }

    const LONG result = ChangeDisplaySettingsEx(
        devices[device].name, &device_modes[mode], NULL,
        mode == devices[device].default_mode ? 0 : CDS_FULLSCREEN, NULL);
    if (result != DISP_CHANGE_SUCCESSFUL) {
        DLOG("Failed to change to mode %d (%dx%dx%d @ %dHz) on device %s:"
             " %ld", mode, (int)device_modes[mode].dmPelsWidth,
             (int)device_modes[mode].dmPelsHeight,
             (int)device_modes[mode].dmBitsPerPel,
             (int)device_modes[mode].dmDisplayFrequency,
             devices[device].name, result);
        return 0;
    }

    current_mode = mode;
    return 1;
}

/*-----------------------------------------------------------------------*/

static void update_mouse_grab(void)
{
    int grab = grab_requested;

    /* Don't grab if there's no window visible or the window doesn't have
     * input focus. */
    if (!current_window || window_minimized || !window_focused) {
        grab = 0;
    }

    if (grab && !mouse_grabbed) {
        RECT rect;
        if (UNLIKELY(!GetClientRect(current_window, &rect))) {
            DLOG("GetClientRect() failed: %s",
                 windows_strerror(GetLastError()));
        } else if (UNLIKELY(!ClipCursor(&rect))) {
            DLOG("ClipCursor() failed: %s", windows_strerror(GetLastError()));
        } else {
            mouse_grabbed = 1;
        }
    } else if (!grab && mouse_grabbed) {
        ClipCursor(NULL);
        mouse_grabbed = 0;
    }
}

/*-----------------------------------------------------------------------*/

static void close_window(void)
{
    PRECOND(current_window, return);

    opengl_cleanup();
    wglMakeCurrent(GetDC(current_window), NULL);
    wglDeleteContext(wgl_context);
    wgl_context = NULL;
    DestroyWindow(current_window);
    current_window = NULL;
    window_focused = 0;

    update_mouse_grab();
    windows_show_mouse_pointer(1);
}

/*-----------------------------------------------------------------------*/

static int setup_opengl(int width, int height)
{
    opengl_enable_debug(use_opengl_debug);

    uint32_t gl_flags = OPENGL_FEATURE_FAST_STATIC_VBO
                      | OPENGL_FEATURE_FAST_DYNAMIC_VBO;
    /* Only use glGenerateMipmap() for drivers claiming to support GL3.0
     * (in which the function became standard), since earlier drivers may
     * have broken implementations. */
    const char *version_string = (const char *)glGetString(GL_VERSION);
    if (version_string && strtoul(version_string, NULL, 10) >= 3) {
        gl_flags |= OPENGL_FEATURE_GENERATEMIPMAP;
    }
    /* We could theoretically use native quads, but all modern GPUs just
     * convert them into triangles anyway, and some OpenGL drivers are
     * broken with respect to quads (e.g. the VMware driver renders three
     * vertices of a QUAD_STRIP as a triangle), so it's not worth it. */

    /* Some Intel drivers (at least build 10.18.15.4256) implement the
     * glProgramUniform functions incorrectly, so that modifying a
     * uniform on a non-current shader makes that shader current.
     * To avoid this, we just disable separate shaders on Intel drivers
     * entirely.  It's low-end hardware anyway, so it's not like users
     * should be expecting good performance in the first place... */
    const char *vendor = (const char *)glGetString(GL_VENDOR);
    if (vendor && strcmp(vendor, "Intel") == 0) {
        gl_flags |= OPENGL_FEATURE_NO_SEPARATE_SHADERS;
    }

    if (UNLIKELY(!opengl_init(width, height, gl_flags))) {
        DLOG("Failed to set up OpenGL!");
        return 0;
    }

    opengl_set_compile_context_callback(create_gl_shader_compilation_context);

    return 1;
}

/*-----------------------------------------------------------------------*/

static int wgl_has_extension(const char *name)
{
    PRECOND(current_window, return 0);

    PFNWGLGETEXTENSIONSSTRINGARBPROC p_wglGetExtensionsStringARB =
        (void *)wglGetProcAddress("wglGetExtensionsStringARB");
    if (p_wglGetExtensionsStringARB) {
        const char *extensions =
            (*p_wglGetExtensionsStringARB)(GetDC(current_window));
        if (extensions) {
            const int namelen = strlen(name);
            const char *s = extensions;
            while ((s = strstr(s, name)) != NULL) {
                if ((s == extensions || s[-1] == ' ')
                 && (s[namelen] == 0 || s[namelen] == ' ')) {
                    return 1;
                }
                s += namelen;
                s += strcspn(s, " ");
            }
        }
    }
    return 0;
}

/*-----------------------------------------------------------------------*/

static void *wglGetProcAddress_wrapper(const char *name)
{
    void *function = wglGetProcAddress(name);
    if (!function) {
        /* wglGetProcAddress() does _not_ return function pointers for
         * OpenGL 1.0/1.1 functions (why?!) so try looking them up via
         * the Windows API. */
        function = GetProcAddress(GetModuleHandle("opengl32.dll"), name);
    }
    return function;
}

/*-----------------------------------------------------------------------*/

static HGLRC create_wgl_context(HDC dc)
{
    HGLRC context = 0;
    if (use_opengl_debug) {
        HGLRC (WINAPI *p_wglCreateContextAttribsARB)(HDC, HGLRC, const int *) =
            (void *)GetProcAddress(GetModuleHandle("opengl32.dll"),
                                   "wglGetContextAttribsARB");
        if (UNLIKELY(!p_wglCreateContextAttribsARB)) {
            DLOG("wglCreateContextAttribsARB() not found, debug information"
                 " may not be available");
        } else {
            context = (*p_wglCreateContextAttribsARB)(
                dc, 0, (const int[]){WGL_CONTEXT_FLAGS_ARB,
                                     WGL_CONTEXT_DEBUG_BIT_ARB,
                                     0});
            if (UNLIKELY(!context)) {
                DLOG("Failed to create a debug context, debug information"
                     " may not be available");
            }
        }
    }
    if (!context) {
        context = wglCreateContext(dc);
    }
    if (UNLIKELY(!context)) {
        DLOG("wglCreateContext() failed: %s",
             windows_strerror(GetLastError()));
        return 0;
    }
    if (UNLIKELY(!wglMakeCurrent(dc, context))) {
        DLOG("wglMakeCurrent() failed: %s", windows_strerror(GetLastError()));
        wglDeleteContext(context);
        return 0;
    }
    return context;
}

/*-----------------------------------------------------------------------*/

static int create_gl_shader_compilation_context(void)
{
    if (wglGetCurrentContext()) {
        return 1;
    }

    if (!current_window) {
        DLOG("No window open, can't create subthread context");
        return 0;
    }
    if (!create_wgl_context(GetDC(current_window))) {
        DLOG("Failed to create subthread context");
        return 0;
    }
    return 1;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
