/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * tools/oggtowav.c: Program to convert Ogg Vorbis audio to linear PCM in
 * RIFF WAVE format.  Requires libvorbisfile for reading Ogg Vorbis files.
 */

/*
 * To use, run this program as:
 *
 *     oggtomp3 file1.ogg [file2.ogg...]
 *
 * This will convert all named Ogg Vorbis files to RIFF WAVE format.
 * The converted files are saved with an extension of .wav replacing any
 * .ogg extension in the input filenames.
 *
 * Note that this program does not detect output files larger than 4GB;
 * such files will have an invalid header, but no error will be reported.
 */

#include "tool-common.h"

#include <vorbis/vorbisfile.h>

/*************************************************************************/

static int convert_to_wav(const char *file);

/*************************************************************************/
/*************************************************************************/

/**
 * main:  Program entry point.  Parses command-line parameters and calls
 * convert_to_wav() for each file specified.
 *
 * [Parameters]
 *     argc: Command line argument count.
 *     argv: Command line argument array.
 * [Return value]
 *     Zero on successful completion, nonzero if an error occurred.
 */
int main(int argc, char **argv)
{
    if (argc < 2 || argv[1][0] == '-') {
        fprintf(stderr, "Usage: %s file1.ogg [file2.ogg...]\n", argv[0]);
        return 1;
    }

    for (int argi = 1; argi < argc; argi++) {
        if (!convert_to_wav(argv[argi])) {
            fprintf(stderr, "Failed to convert %s\n", argv[argi]);
            return 1;
        }
    }

    return 0;
}

/*-----------------------------------------------------------------------*/

/**
 * convert_to_wav:  Convert an Ogg Vorbis audio file to PCM in RIFF WAVE.
 *
 * [Parameters]
 *     path: Pathname of Ogg Vorbis file to convert.
 * [Return value]
 *     True on success, false on error.
 */
static int convert_to_wav(const char *path)
{
    /* Generate the WAV output filename. */

    char wav_pathbuf[1000];
    unsigned int pathlen =
        snprintf(wav_pathbuf, sizeof(wav_pathbuf)-4, "%s", path);
    if (pathlen >= sizeof(wav_pathbuf)-4) {
        DLOG("Pathname too long: %s", path);
        goto error_return;
    }
    if (stricmp(wav_pathbuf + pathlen - 4, ".ogg") == 0) {
        wav_pathbuf[pathlen-4] = 0;
        pathlen -= 4;
    }
    memcpy(wav_pathbuf + pathlen, ".wav", 5);

    /* Open the input and output files. */

    FILE *in = fopen(path, "rb");
    if (!in) {
        perror(path);
        goto error_return;
    }
    FILE *out = fopen(wav_pathbuf, "wb");
    if (!out) {
        perror(wav_pathbuf);
        goto error_close_in;
    }

    /* Initialize an Ogg Vorbis reader instance, and obtain the source
     * audio parameters. */

    OggVorbis_File vf;
    if (ov_open_callbacks(in, &vf, NULL, 0, OV_CALLBACKS_NOCLOSE) != 0) {
        fprintf(stderr, "%s: ov_open() failed\n", path);
        goto error_close_out;
    }

    vorbis_info *info = ov_info(&vf, -1);
    if (!info) {
        fprintf(stderr, "%s: ov_info() failed\n", path);
        goto error_close_vorbis;
    }
    if (info->channels != 1 && info->channels != 2) {
        fprintf(stderr, "%s: Bad channel count %d\n", path, info->channels);
        goto error_close_vorbis;
    }

    /* Read the PCM data into memory. */

    uint32_t total_size = 0;
    uint8_t *pcm_buffer = NULL;
    uint32_t buffer_size = 0;
    for (;;) {
        if (total_size >= buffer_size) {
            buffer_size += 1000000;
            uint8_t *new_buffer = realloc(pcm_buffer, buffer_size);
            if (!new_buffer) {
                fprintf(stderr, "%s: Out of memory for PCM data (%u bytes)\n",
                        path, buffer_size);
                goto error_free_pcm_buffer;
            }
            pcm_buffer = new_buffer;
        }

        int bitstream_unused;
        const int32_t nread = ov_read(
            &vf, (char *)(pcm_buffer + total_size), buffer_size - total_size,
            /*bigendianp*/ 0, /*word*/ 2, /*sgned*/ 1, &bitstream_unused
        );
        if (nread == 0 || nread == OV_EOF) {
            break;
        } else if (nread == OV_HOLE) {
            fprintf(stderr, "%s: Warning: Possible corrupt data at sample"
                    " %u, continuing anyway\n", path,
                    total_size / (info->channels*2));
        } else if (nread < 0) {
            fprintf(stderr, "%s: Error %d decompressing Ogg Vorbis data\n",
                    path, nread);
            goto error_free_pcm_buffer;
        } else {
            total_size += nread;
        }
    }

    /* Construct a RIFF WAVE header for the file, then write it and the
     * PCM data to the output file. */

    struct {
        char RIFF[4];
        uint8_t filesize[4];
        char WAVE[4];
        char fmt_[4];
        uint8_t fmt_size[4];
        struct {
            uint8_t codec[2];
            uint8_t channels[2];
            uint8_t srate[4];
            uint8_t datarate[4];
            uint8_t samplesize[2];
            uint8_t bits[2];
        } fmt_struct;
        char data[4];
        uint8_t datasize[4];
    } header;
    memcpy(header.RIFF, "RIFF", 4);
    header.filesize[0] = (total_size + sizeof(header)-8) >>  0;
    header.filesize[1] = (total_size + sizeof(header)-8) >>  8;
    header.filesize[2] = (total_size + sizeof(header)-8) >> 16;
    header.filesize[3] = (total_size + sizeof(header)-8) >> 24;
    memcpy(header.WAVE, "WAVE", 4);
    memcpy(header.fmt_, "fmt ", 4);
    header.fmt_size[0] = sizeof(header.fmt_struct) >>  0;
    header.fmt_size[1] = sizeof(header.fmt_struct) >>  8;
    header.fmt_size[2] = sizeof(header.fmt_struct) >> 16;
    header.fmt_size[3] = sizeof(header.fmt_struct) >> 24;
    header.fmt_struct.codec[0] = 0x0001 >> 0;
    header.fmt_struct.codec[1] = 0x0001 >> 8;
    header.fmt_struct.channels[0] = info->channels >> 0;
    header.fmt_struct.channels[1] = info->channels >> 8;
    header.fmt_struct.srate[0] = info->rate >>  0;
    header.fmt_struct.srate[1] = info->rate >>  8;
    header.fmt_struct.srate[2] = info->rate >> 16;
    header.fmt_struct.srate[3] = info->rate >> 24;
    header.fmt_struct.datarate[0] = (info->rate * info->channels * 2) >>  0;
    header.fmt_struct.datarate[1] = (info->rate * info->channels * 2) >>  8;
    header.fmt_struct.datarate[2] = (info->rate * info->channels * 2) >> 16;
    header.fmt_struct.datarate[3] = (info->rate * info->channels * 2) >> 24;
    header.fmt_struct.samplesize[0] = (info->channels * 2) >> 0;
    header.fmt_struct.samplesize[1] = (info->channels * 2) >> 8;
    header.fmt_struct.bits[0] = 16 >> 0;
    header.fmt_struct.bits[1] = 16 >> 8;
    memcpy(header.data, "data", 4);
    header.datasize[0] = total_size >>  0;
    header.datasize[1] = total_size >>  8;
    header.datasize[2] = total_size >> 16;
    header.datasize[3] = total_size >> 24;

    if (fwrite(&header, sizeof(header), 1, out) != 1) {
        fprintf(stderr, "%s: Write error on RIFF WAVE header\n", wav_pathbuf);
        goto error_free_pcm_buffer;
    }
    if (fwrite(pcm_buffer, 1, total_size, out) != total_size) {
        fprintf(stderr, "%s: Write error on PCM data\n", wav_pathbuf);
        goto error_free_pcm_buffer;
    }

    /* All done! */

    free(pcm_buffer);
    ov_clear(&vf);
    fclose(out);
    fclose(in);
    return 1;

    /* Error handling. */

  error_free_pcm_buffer:
    free(pcm_buffer);
  error_close_vorbis:
    ov_clear(&vf);
  error_close_out:
    fclose(out);
    remove(wav_pathbuf);
  error_close_in:
    fclose(in);
  error_return:
    return 0;

}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
